@@ -23,50 +23,50 @@
 
 ;;; Commentary:
 
-;; Access functions for crypted remote files.  It uses encfs to
+;; Access functions for encrypted remote files.  It uses encfs to
 ;; encrypt / decrypt the files on a remote directory.  A remote
-;; directory, which shall include crypted files, must be declared in
+;; directory where you wish files to be encrypted must be declared in
 ;; `tramp-crypt-directories' via command `tramp-crypt-add-directory'.
 ;; All files in that directory, including all subdirectories, are
-;; stored there encrypted.  This includes file names and directory
+;; stored encrypted.  This includes file names and directory
 ;; names.
 
 ;; This package is just responsible for the encryption part.  Copying
-;; of the crypted files is still the responsibility of the remote file
-;; name handlers.
+;; of the encrypted files is still the responsibility of the remote
+;; file name handlers.
 
 ;; A password protected encfs configuration file is created the very
-;; first time you access a crypted remote directory.  It is kept in
-;; your user directory "~/.emacs.d/" with the url-encoded directory
-;; name as part of the basename, and ".encfs6.xml" as suffix.  Do not
-;; lose this file and the corresponding password; otherwise there is
-;; no way to decrypt your crypted files.
+;; first time you access an encrypted remote directory.  It is kept in
+;; your user directory (usually "~/.emacs.d/") with the url-encoded
+;; directory name as part of the basename, and ".encfs6.xml" as
+;; suffix.  Do not lose this file and the corresponding password;
+;; otherwise there is no way to decrypt your encrypted files.
 
 ;; If the user option `tramp-crypt-save-encfs-config-remote' is
 ;; non-nil (the default), the encfs configuration file ".encfs6.xml"
-;; is also kept in the crypted remote directory.  It depends on you,
+;; is also kept in the encrypted remote directory.  It depends on you,
 ;; whether you regard the password protection of this file as
-;; sufficient.
+;; sufficient security.
 
 ;; If you use a remote file name with a quoted localname part, this
 ;; localname and the corresponding file will not be encrypted/
-;; decrypted.  For example, if you have a crypted remote directory
-;; "/nextcloud:user@host:/crypted_dir", the command
+;; decrypted.  For example, if you have an encrypted remote directory
+;; "/nextcloud:user@host:/encrypted_dir", the command
 ;;
-;;   C-x d /nextcloud:user@host:/crypted_dir
+;;   C-x d /nextcloud:user@host:/encrypted_dir
 ;;
 ;; will show the directory listing with the plain file names, and the
 ;; command
 ;;
-;;   C-x d /nextcloud:user@host:/:/crypted_dir
+;;   C-x d /nextcloud:user@host:/:/encrypted_dir
 ;;
 ;; will show the directory with the encrypted file names, and visiting
-;; a file will show its crypted contents.  However, it is highly
-;; discouraged to mix crypted and not crypted files in the same
+;; a file will show its encrypted contents.  However, it is highly
+;; discouraged to mix encrypted and non-encrypted files in the same
 ;; directory.
 
-;; If a remote directory shall not include crypted files anymore, it
-;; must be indicated by the command `tramp-crypt-remove-directory'.
+;; To disable encryption for a particular remote directory, use the
+;; command `tramp-crypt-remove-directory'.
 
 ;;; Code:
 
@@ -78,7 +78,7 @@
 (autoload 'text-property-search-forward "text-property-search")
 
 (defconst tramp-crypt-method "crypt"
-  "Method name for crypted remote directories.")
+  "Method name for encrypted remote directories.")
 
 (defcustom tramp-crypt-encfs-program "encfs"
   "Name of the encfs program."
@@ -96,7 +96,7 @@
   "Configuration option for encfs.
 This could be either \"--standard\" or \"--paranoia\".  The file
 name IV chaining mode mode will always be disabled when
-initializing a new crypted remote directory."
+initializing a new encrypted remote directory."
   :group 'tramp
   :version "28.1"
   :type '(choice (const "--standard")
@@ -120,7 +120,7 @@ initializing a new crypted remote directory."
 They are completed by \"M-x TAB\" only when encryption support is enabled."
   (and tramp-crypt-enabled
        ;; `tramp-crypt-remove-directory' needs to be completed only in
-       ;; case we have already crypted directories.
+       ;; case we have already encrypted directories.
        (or (not (eq symbol #'tramp-crypt-remove-directory))
 	   tramp-crypt-directories)))
 
@@ -129,21 +129,21 @@ They are completed by \"M-x TAB\" only when encryption support is enabled."
   "Encfs configuration file name.")
 
 (defcustom tramp-crypt-save-encfs-config-remote t
-  "Whether to keep the encfs configuration file in the crypted remote directory."
+  "Whether to keep the encfs configuration file in the encrypted remote directory."
   :group 'tramp
   :version "28.1"
   :type 'boolean)
 
 ;;;###tramp-autoload
 (defvar tramp-crypt-directories nil
-  "List of crypted remote directories.")
+  "List of encrypted remote directories.")
 
 ;; It must be a `defsubst' in order to push the whole code into
 ;; tramp-loaddefs.el.  Otherwise, there would be recursive autoloading.
 ;;;###tramp-autoload
 (defsubst tramp-crypt-file-name-p (name)
-  "Return the crypted remote directory NAME belongs to.
-If NAME doesn't belong to a crypted remote directory, retun nil."
+  "Return the encrypted remote directory NAME belongs to.
+If NAME doesn't belong to an encrypted remote directory, return nil."
   (catch 'crypt-file-name-p
     (and tramp-crypt-enabled (stringp name)
 	 (not (tramp-compat-file-name-quoted-p name))
@@ -151,13 +151,14 @@ If NAME doesn't belong to a crypted remote directory, retun nil."
 	 (dolist (dir tramp-crypt-directories)
 	   (and (string-prefix-p
 		 dir (file-name-as-directory (expand-file-name name)))
-		(throw  'crypt-file-name-p dir))))))
+		(throw 'crypt-file-name-p dir))))))
 
 
 ;; New handlers should be added here.
 ;;;###tramp-autoload
 (defconst tramp-crypt-file-name-handler-alist
-  '((access-file . tramp-crypt-handle-access-file)
+  '(;; `abbreviate-file-name' performed by default handler.
+    (access-file . tramp-crypt-handle-access-file)
     (add-name-to-file . tramp-handle-add-name-to-file)
     ;; `byte-compiler-base-file-name' performed by default handler.
     (copy-directory . tramp-handle-copy-directory)
@@ -208,6 +209,7 @@ If NAME doesn't belong to a crypted remote directory, retun nil."
     ;; `get-file-buffer' performed by default handler.
     (insert-directory . tramp-crypt-handle-insert-directory)
     (insert-file-contents . tramp-handle-insert-file-contents)
+    (list-system-processes . ignore)
     (load . tramp-handle-load)
     (lock-file . tramp-crypt-handle-lock-file)
     (make-auto-save-file-name . tramp-handle-make-auto-save-file-name)
@@ -217,6 +219,8 @@ If NAME doesn't belong to a crypted remote directory, retun nil."
     (make-nearby-temp-file . tramp-handle-make-nearby-temp-file)
     (make-process . ignore)
     (make-symbolic-link . tramp-handle-make-symbolic-link)
+    (memory-info . ignore)
+    (process-attributes . ignore)
     (process-file . ignore)
     (rename-file . tramp-crypt-handle-rename-file)
     (set-file-acl . ignore)
@@ -228,7 +232,9 @@ If NAME doesn't belong to a crypted remote directory, retun nil."
     (start-file-process . ignore)
     ;; `substitute-in-file-name' performed by default handler.
     (temporary-file-directory . tramp-handle-temporary-file-directory)
+    ;; `tramp-get-home-directory' performed by default-handler.
     ;; `tramp-get-remote-gid' performed by default handler.
+    ;; `tramp-get-remote-groups' performed by default handler.
     ;; `tramp-get-remote-uid' performed by default handler.
     (tramp-set-file-uid-gid . tramp-crypt-handle-set-file-uid-gid)
     (unhandled-file-name-directory . ignore)
@@ -240,7 +246,7 @@ If NAME doesn't belong to a crypted remote directory, retun nil."
 Operations not mentioned here will be handled by the default Emacs primitives.")
 
 (defsubst tramp-crypt-file-name-for-operation (operation &rest args)
-  "Like `tramp-file-name-for-operation', but for crypted remote files."
+  "Like `tramp-file-name-for-operation', but for encrypted remote files."
   (let ((tfnfo (apply #'tramp-file-name-for-operation operation args)))
     ;; `tramp-file-name-for-operation' returns already the first argument
     ;; if it is remote.  So we check a possible second argument.
@@ -264,7 +270,7 @@ arguments to pass to the OPERATION."
 
 ;;;###tramp-autoload
 (defun tramp-crypt-file-name-handler (operation &rest args)
-  "Invoke the crypted remote file related OPERATION.
+  "Invoke the encrypted remote file related OPERATION.
 First arg specifies the OPERATION, second arg is a list of
 arguments to pass to the OPERATION."
   (if-let ((filename
@@ -294,8 +300,8 @@ arguments to pass to the OPERATION."
 (defun tramp-crypt-config-file-name (vec)
   "Return the encfs config file name for VEC."
   (expand-file-name
-   (concat "tramp-" (tramp-file-name-host vec) tramp-crypt-encfs-config)
-   user-emacs-directory))
+   (locate-user-emacs-file
+    (concat "tramp-" (tramp-file-name-host vec) tramp-crypt-encfs-config))))
 
 (defun tramp-crypt-maybe-open-connection (vec)
   "Maybe open a connection VEC.
@@ -312,7 +318,7 @@ connection if a previous connection has died for some reason."
       (process-put p 'vector vec)
       (set-process-query-on-exit-flag p nil)))
 
-  ;; The following operations must be performed w/o
+  ;; The following operations must be performed without
   ;; `tramp-crypt-file-name-handler'.
   (let* (tramp-crypt-enabled
 	 ;; Don't check for a proper method.
@@ -322,7 +328,7 @@ connection if a previous connection has died for some reason."
 	   tramp-crypt-encfs-config (tramp-crypt-get-remote-dir vec)))
 	 (local-config (tramp-crypt-config-file-name vec)))
     ;; There is no local encfs6 config file.
-    (when (not (file-exists-p local-config))
+    (unless (file-exists-p local-config)
       (if (and tramp-crypt-save-encfs-config-remote
 	       (file-exists-p remote-config))
 	  ;; Copy remote encfs6 config file if possible.
@@ -409,7 +415,7 @@ ARGS are the arguments.  It returns t if ran successful, and nil otherwise."
 	t))))
 
 (defun tramp-crypt-do-encrypt-or-decrypt-file-name (op name)
-  "Return encrypted / decrypted NAME if NAME belongs to a crypted directory.
+  "Return encrypted / decrypted NAME if NAME belongs to an encrypted directory.
 OP must be `encrypt' or `decrypt'.  Raise an error if this fails.
 Otherwise, return NAME."
   (if-let ((tramp-crypt-enabled t)
@@ -422,7 +428,7 @@ Otherwise, return NAME."
        (if (directory-name-p name) #'file-name-as-directory #'identity)
        (concat
 	dir
-	(unless (string-equal localname "/")
+	(unless (string-match-p (rx bos (? "/") eos) localname)
 	  (with-tramp-file-property
 	      crypt-vec localname (concat (symbol-name op) "-file-name")
 	    (unless (tramp-crypt-send-command
@@ -433,24 +439,24 @@ Otherwise, return NAME."
 	       (if (eq op 'encrypt) "Encoding" "Decoding") name))
 	    (with-current-buffer (tramp-get-connection-buffer crypt-vec)
 	      (goto-char (point-min))
-	      (buffer-substring (point-min) (point-at-eol)))))))
+	      (buffer-substring (point-min) (line-end-position)))))))
     ;; Nothing to do.
     name))
 
 (defsubst tramp-crypt-encrypt-file-name (name)
-  "Return encrypted NAME if NAME belongs to a crypted directory.
+  "Return encrypted NAME if NAME belongs to an encrypted directory.
 Otherwise, return NAME."
   (tramp-crypt-do-encrypt-or-decrypt-file-name 'encrypt name))
 
 (defsubst tramp-crypt-decrypt-file-name (name)
-  "Return decrypted NAME if NAME belongs to a crypted directory.
+  "Return decrypted NAME if NAME belongs to an encrypted directory.
 Otherwise, return NAME."
   (tramp-crypt-do-encrypt-or-decrypt-file-name 'decrypt name))
 
 (defun tramp-crypt-do-encrypt-or-decrypt-file (op root infile outfile)
-  "Encrypt / decrypt file INFILE to OUTFILE according to crypted directory ROOT.
+  "Encrypt / decrypt file INFILE to OUTFILE according to encrypted directory ROOT.
 Both files must be local files.  OP must be `encrypt' or `decrypt'.
-If OP ist `decrypt', the basename of INFILE must be an encrypted file name.
+If OP is `decrypt', the basename of INFILE must be an encrypted file name.
 Raise an error if this fails."
   (when-let ((tramp-crypt-enabled t)
 	     (dir (tramp-crypt-file-name-p root))
@@ -470,12 +476,12 @@ Raise an error if this fails."
 	(write-region nil nil outfile)))))
 
 (defsubst tramp-crypt-encrypt-file (root infile outfile)
-  "Encrypt file INFILE to OUTFILE according to crypted directory ROOT.
+  "Encrypt file INFILE to OUTFILE according to encrypted directory ROOT.
 See `tramp-crypt-do-encrypt-or-decrypt-file'."
   (tramp-crypt-do-encrypt-or-decrypt-file 'encrypt root infile outfile))
 
 (defsubst tramp-crypt-decrypt-file (root infile outfile)
-  "Decrypt file INFILE to OUTFILE according to crypted directory ROOT.
+  "Decrypt file INFILE to OUTFILE according to encrypted directory ROOT.
 See `tramp-crypt-do-encrypt-or-decrypt-file'."
   (tramp-crypt-do-encrypt-or-decrypt-file 'decrypt root infile outfile))
 
@@ -485,6 +491,7 @@ See `tramp-crypt-do-encrypt-or-decrypt-file'."
 Files in that directory and all subdirectories will be encrypted
 before copying to, and decrypted after copying from that
 directory.  File names will be also encrypted."
+  ;; (declare (completion tramp-crypt-command-completion-p))
   (interactive "DRemote directory name: ")
   (unless tramp-crypt-enabled
     (tramp-user-error nil "Feature is not enabled."))
@@ -537,10 +544,10 @@ localname."
 	(make-tramp-file-name
 	 :method tramp-crypt-method :user (user-login-name)
 	 :host (url-hexify-string dir))
-      (tramp-user-error nil "Not a crypted remote directory: \"%s\"" name))))
+      (tramp-user-error nil "Not an encrypted remote directory: \"%s\"" name))))
 
 (defun tramp-crypt-get-remote-dir (vec)
-  "Return the name of the crypted remote directory to be used for encfs."
+  "Return the name of the encrypted remote directory to be used for encfs."
   (url-unhex-string (tramp-file-name-host vec)))
 
 
@@ -549,7 +556,7 @@ localname."
 (defun tramp-crypt-handle-access-file (filename string)
   "Like `access-file' for Tramp files."
   (let* ((encrypt-filename (tramp-crypt-encrypt-file-name filename))
-	 (encrypt-regexp (concat (regexp-quote encrypt-filename) "\\'"))
+	 (encrypt-regexp (tramp-compat-rx (literal encrypt-filename) eos))
 	 tramp-crypt-enabled)
     (condition-case err
 	(access-file encrypt-filename string)
@@ -578,6 +585,7 @@ This function is invoked by `tramp-crypt-handle-copy-file' and
 `tramp-crypt-handle-rename-file'.  It is an error if OP is
 neither of `copy' and `rename'.  FILENAME and NEWNAME must be
 absolute file names."
+  ;; FILENAME and NEWNAME are already expanded.
   (unless (memq op '(copy rename))
     (error "Unknown operation `%s', must be `copy' or `rename'" op))
 
@@ -595,62 +603,61 @@ absolute file names."
 	    (delete-directory filename 'recursive)))
 
       (with-parsed-tramp-file-name (if t1 filename newname) nil
-	(unless (file-exists-p filename)
-	  (tramp-compat-file-missing v filename))
-	(when (and (not ok-if-already-exists) (file-exists-p newname))
-	  (tramp-error v 'file-already-exists newname))
-	(when (and (file-directory-p newname)
-		   (not (directory-name-p newname)))
-	  (tramp-error v 'file-error "File is a directory %s" newname))
+	(tramp-barf-if-file-missing v filename
+	  (when (and (not ok-if-already-exists) (file-exists-p newname))
+	    (tramp-error v 'file-already-exists newname))
+	  (when (and (file-directory-p newname)
+		     (not (directory-name-p newname)))
+	    (tramp-error v 'file-error "File is a directory %s" newname))
 
-	(with-tramp-progress-reporter
-	    v 0 (format "%s %s to %s" msg-operation filename newname)
-	  (if (and t1 t2 (string-equal t1 t2))
-	      ;; Both files are on the same crypted remote directory.
-	      (let (tramp-crypt-enabled)
-		(if (eq op 'copy)
-		    (copy-file
-		     encrypt-filename encrypt-newname ok-if-already-exists
-		     keep-date preserve-uid-gid preserve-extended-attributes)
-		  (rename-file
-		   encrypt-filename encrypt-newname ok-if-already-exists)))
+	  (with-tramp-progress-reporter
+	      v 0 (format "%s %s to %s" msg-operation filename newname)
+	    (if (and t1 t2 (string-equal t1 t2))
+		;; Both files are on the same encrypted remote directory.
+		(let (tramp-crypt-enabled)
+		  (if (eq op 'copy)
+		      (copy-file
+		       encrypt-filename encrypt-newname ok-if-already-exists
+		       keep-date preserve-uid-gid preserve-extended-attributes)
+		    (rename-file
+		     encrypt-filename encrypt-newname ok-if-already-exists)))
 
-	    (let* ((tmpdir (tramp-compat-make-temp-file filename 'dir))
-		   (tmpfile1
-		    (expand-file-name
-		     (file-name-nondirectory encrypt-filename) tmpdir))
-		   (tmpfile2
-		    (expand-file-name
-		     (file-name-nondirectory encrypt-newname) tmpdir))
-		   tramp-crypt-enabled)
-	      (cond
-	       ;; Source and target file are on a crypted remote directory.
-	       ((and t1 t2)
-		(if (eq op 'copy)
-		    (copy-file
-		     encrypt-filename encrypt-newname ok-if-already-exists
-		     keep-date preserve-uid-gid preserve-extended-attributes)
-		  (rename-file
-		   encrypt-filename encrypt-newname ok-if-already-exists)))
-	       ;; Source file is on a crypted remote directory.
-	       (t1
-		(if (eq op 'copy)
-		    (copy-file
-		     encrypt-filename tmpfile1 t keep-date preserve-uid-gid
-		     preserve-extended-attributes)
-		  (rename-file encrypt-filename tmpfile1 t))
-		(tramp-crypt-decrypt-file t1 tmpfile1 tmpfile2)
-		(rename-file tmpfile2 newname ok-if-already-exists))
-	       ;; Target file is on a crypted remote directory.
-	       (t2
-		(if (eq op 'copy)
-		    (copy-file
-		     filename tmpfile1 t keep-date preserve-uid-gid
-		     preserve-extended-attributes)
-		  (rename-file filename tmpfile1 t))
-		(tramp-crypt-encrypt-file t2 tmpfile1 tmpfile2)
-		(rename-file tmpfile2 encrypt-newname ok-if-already-exists)))
-	      (delete-directory tmpdir 'recursive))))))
+	      (let* ((tmpdir (tramp-compat-make-temp-file filename 'dir))
+		     (tmpfile1
+		      (expand-file-name
+		       (file-name-nondirectory encrypt-filename) tmpdir))
+		     (tmpfile2
+		      (expand-file-name
+		       (file-name-nondirectory encrypt-newname) tmpdir))
+		     tramp-crypt-enabled)
+		(cond
+		 ;; Source and target file are on an encrypted remote directory.
+		 ((and t1 t2)
+		  (if (eq op 'copy)
+		      (copy-file
+		       encrypt-filename encrypt-newname ok-if-already-exists
+		       keep-date preserve-uid-gid preserve-extended-attributes)
+		    (rename-file
+		     encrypt-filename encrypt-newname ok-if-already-exists)))
+		 ;; Source file is on an encrypted remote directory.
+		 (t1
+		  (if (eq op 'copy)
+		      (copy-file
+		       encrypt-filename tmpfile1 t keep-date preserve-uid-gid
+		       preserve-extended-attributes)
+		    (rename-file encrypt-filename tmpfile1 t))
+		  (tramp-crypt-decrypt-file t1 tmpfile1 tmpfile2)
+		  (rename-file tmpfile2 newname ok-if-already-exists))
+		 ;; Target file is on an encrypted remote directory.
+		 (t2
+		  (if (eq op 'copy)
+		      (copy-file
+		       filename tmpfile1 t keep-date preserve-uid-gid
+		       preserve-extended-attributes)
+		    (rename-file filename tmpfile1 t))
+		  (tramp-crypt-encrypt-file t2 tmpfile1 tmpfile2)
+		  (rename-file tmpfile2 encrypt-newname ok-if-already-exists)))
+		(delete-directory tmpdir 'recursive)))))))
 
     (when (and t1 (eq op 'rename))
       (with-parsed-tramp-file-name filename v1
@@ -677,7 +684,7 @@ absolute file names."
      (list filename newname ok-if-already-exists keep-date
 	   preserve-uid-gid preserve-extended-attributes))))
 
-;; Crypted files won't be trashed.
+;; Encrypted files won't be trashed.
 (defun tramp-crypt-handle-delete-directory
     (directory &optional recursive _trash)
   "Like `delete-directory' for Tramp files."
@@ -686,7 +693,7 @@ absolute file names."
     (let (tramp-crypt-enabled)
       (delete-directory (tramp-crypt-encrypt-file-name directory) recursive))))
 
-;; Crypted files won't be trashed.
+;; Encrypted files won't be trashed.
 (defun tramp-crypt-handle-delete-file (filename &optional _trash)
   "Like `delete-file' for Tramp files."
   (with-parsed-tramp-file-name (expand-file-name filename) nil
@@ -697,36 +704,14 @@ absolute file names."
 (defun tramp-crypt-handle-directory-files
     (directory &optional full match nosort count)
   "Like `directory-files' for Tramp files."
-  (unless (file-exists-p directory)
-    (tramp-compat-file-missing (tramp-dissect-file-name directory) directory))
-  (when (file-directory-p directory)
-    (setq directory (file-name-as-directory (expand-file-name directory)))
-    (let* (tramp-crypt-enabled
-	   (result
-	    (directory-files (tramp-crypt-encrypt-file-name directory) 'full)))
-      (setq result
-	    (mapcar (lambda (x) (tramp-crypt-decrypt-file-name x)) result))
-      (when match
-	(setq result
-	      (delq
-	       nil
-	       (mapcar
-		(lambda (x)
-		  (when (string-match-p match (substring x (length directory)))
-		    x))
-		result))))
-      (unless full
-	(setq result
-	      (mapcar
-	       (lambda (x)
-		 (replace-regexp-in-string
-		  (concat "^" (regexp-quote directory)) "" x))
-	       result)))
-      (unless nosort
-        (setq result (sort result #'string<)))
-      (when (and (natnump count) (> count 0))
-	(setq result (nbutlast result (- (length result) count))))
-      result)))
+  (tramp-skeleton-directory-files directory full match nosort count
+    (let (tramp-crypt-enabled)
+      (mapcar
+       (lambda (x)
+	 (replace-regexp-in-string
+	  (tramp-compat-rx bos (literal directory)) ""
+	  (tramp-crypt-decrypt-file-name x)))
+       (directory-files (tramp-crypt-encrypt-file-name directory) 'full)))))
 
 (defun tramp-crypt-handle-file-attributes (filename &optional id-format)
   "Like `file-attributes' for Tramp files."
@@ -842,24 +827,21 @@ WILDCARD is not supported."
 
 (defun tramp-crypt-handle-set-file-modes (filename mode &optional flag)
   "Like `set-file-modes' for Tramp files."
-  (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-properties v localname)
+  (tramp-skeleton-set-file-modes-times-uid-gid filename
     (let (tramp-crypt-enabled)
       (tramp-compat-set-file-modes
        (tramp-crypt-encrypt-file-name filename) mode flag))))
 
 (defun tramp-crypt-handle-set-file-times (filename &optional time flag)
   "Like `set-file-times' for Tramp files."
-  (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-properties v localname)
+  (tramp-skeleton-set-file-modes-times-uid-gid filename
     (let (tramp-crypt-enabled)
       (tramp-compat-set-file-times
        (tramp-crypt-encrypt-file-name filename) time flag))))
 
 (defun tramp-crypt-handle-set-file-uid-gid (filename &optional uid gid)
   "Like `tramp-set-file-uid-gid' for Tramp files."
-  (with-parsed-tramp-file-name filename nil
-    (tramp-flush-file-properties v localname)
+  (tramp-skeleton-set-file-modes-times-uid-gid filename
     (let (tramp-crypt-enabled)
       (tramp-set-file-uid-gid
        (tramp-crypt-encrypt-file-name filename) uid gid))))
@@ -871,6 +853,14 @@ WILDCARD is not supported."
     (tramp-compat-funcall
      'unlock-file (tramp-crypt-encrypt-file-name filename))))
 
+(with-eval-after-load 'bookmark
+  (add-hook 'bookmark-inhibit-context-functions
+	    #'tramp-crypt-file-name-p)
+  (add-hook 'tramp-crypt-unload-hook
+	    (lambda ()
+	      (remove-hook 'bookmark-inhibit-context-functions
+			   #'tramp-crypt-file-name-p))))
+
 (add-hook 'tramp-unload-hook
 	  (lambda ()
 	    (unload-feature 'tramp-crypt 'force)))
