@@ -31,7 +31,6 @@
 (require 'ert)
 (require 'ert-x)
 (require 'tramp-archive)
-(defvar tramp-copy-size-limit)
 (defvar tramp-persistency-file-name)
 
 ;; `ert-resource-file' was introduced in Emacs 28.1.
@@ -41,7 +40,8 @@
       "Format for `ert-resource-directory'.")
     (defvar ert-resource-directory-trim-left-regexp ""
       "Regexp for `string-trim' (left) used by `ert-resource-directory'.")
-    (defvar ert-resource-directory-trim-right-regexp "\\(-tests?\\)?\\.el"
+    (defvar ert-resource-directory-trim-right-regexp
+      (rx (? "-test" (? "s")) ".el")
       "Regexp for `string-trim' (right) used by `ert-resource-directory'.")
 
     (defmacro ert-resource-directory ()
@@ -96,7 +96,6 @@ Do not hexlify \"/\".  This hexlified string is used in `file:///' URLs."
 
 (setq password-cache-expiry nil
       tramp-cache-read-persistent-data t ;; For auth-sources.
-      tramp-copy-size-limit nil
       tramp-persistency-file-name nil
       tramp-verbose 0)
 
@@ -122,15 +121,9 @@ the origin of the temporary TMPFILE, have no write permissions."
      (directory-files tmpfile 'full directory-files-no-dot-files-regexp))
     (delete-directory tmpfile)))
 
-(defun tramp-archive--test-emacs26-p ()
-  "Check for Emacs version >= 26.1.
-Some semantics has been changed for there, w/o new functions or
-variables, so we check the Emacs version directly."
-  (>= emacs-major-version 26))
-
 (defun tramp-archive--test-emacs27-p ()
   "Check for Emacs version >= 27.1.
-Some semantics has been changed for there, w/o new functions or
+Some semantics has been changed for there, without new functions or
 variables, so we check the Emacs version directly."
   (>= emacs-major-version 27))
 
@@ -433,7 +426,7 @@ This checks also `file-name-as-directory', `file-name-directory',
 	     (setq tmp-name
 		   (file-local-copy
 		    (expand-file-name "what" tramp-archive-test-archive)))
-	     :type tramp-file-missing))
+	     :type 'file-missing))
 
       ;; Cleanup.
       (ignore-errors (tramp-archive--test-delete tmp-name))
@@ -461,7 +454,7 @@ This checks also `file-name-as-directory', `file-name-directory',
 	  (should-error
 	   (insert-file-contents
 	    (expand-file-name "what" tramp-archive-test-archive))
-	   :type tramp-file-missing))
+	   :type 'file-missing))
 
 	;; Cleanup.
 	(tramp-archive-cleanup-hash))))
@@ -552,11 +545,9 @@ This checks also `file-name-as-directory', `file-name-directory',
 	  (should (file-directory-p tmp-name2))
 	  (should (file-exists-p tmp-name4))
 	  ;; Target directory does exist already.
-	  ;; This has been changed in Emacs 26.1.
-	  (when (tramp-archive--test-emacs26-p)
-	    (should-error
-	     (copy-directory tmp-name1 tmp-name2)
-	     :type 'file-error))
+	  (should-error
+	   (copy-directory tmp-name1 tmp-name2)
+	   :type 'file-error)
 	  (tramp-archive--test-delete tmp-name4)
 	  (copy-directory tmp-name1 (file-name-as-directory tmp-name2))
 	  (should (file-directory-p tmp-name3))
@@ -621,19 +612,19 @@ This checks also `file-name-as-directory', `file-name-directory',
 	 (append '("LANG=C" "LANGUAGE=C" "LC_ALL=C") process-environment)))
     (unwind-protect
 	(progn
-	  ;; Due to Bug#29423, this works only since for Emacs 26.1.
-	  (when nil ;; TODO (tramp-archive--test-emacs26-p)
-	    (with-temp-buffer
-	      (insert-directory tramp-archive-test-archive nil)
-	      (goto-char (point-min))
-	      (should
-	       (looking-at-p (regexp-quote tramp-archive-test-archive)))))
+	  (with-temp-buffer
+	    (insert-directory tramp-archive-test-archive nil)
+	    (goto-char (point-min))
+	    (should
+	     (looking-at-p
+	      (tramp-compat-rx (literal tramp-archive-test-archive)))))
 	  (with-temp-buffer
 	    (insert-directory tramp-archive-test-archive "-al")
 	    (goto-char (point-min))
 	    (should
 	     (looking-at-p
-	      (format "^.+ %s$" (regexp-quote tramp-archive-test-archive)))))
+	      (tramp-compat-rx
+	       bol (+ nonl) blank (literal tramp-archive-test-archive) eol))))
 	  (with-temp-buffer
 	    (insert-directory
 	     (file-name-as-directory tramp-archive-test-archive)
@@ -641,21 +632,23 @@ This checks also `file-name-as-directory', `file-name-directory',
 	    (goto-char (point-min))
 	    (should
 	     (looking-at-p
-	      (concat
-	       ;; There might be a summary line.
-	       "\\(total.+[[:digit:]]+ ?[kKMGTPEZY]?i?B?\n\\)?"
-	       ;; We don't know in which order the files appear.
-	       (format
-		"\\(.+ %s\\( ->.+\\)?\n\\)\\{%d\\}"
-		(regexp-opt (directory-files tramp-archive-test-archive))
-		(length (directory-files tramp-archive-test-archive)))))))
-
+	      (rx-to-string
+	       `(:
+		 ;; There might be a summary line.
+		 (? "total" (+ nonl) (+ digit) (? blank)
+		    (? (any "EGKMPTYZk")) (? "i") (? "B") "\n")
+		 ;; We don't know in which order the files appear.
+		 (= ,(length (directory-files tramp-archive-test-archive))
+		    (+ nonl) blank
+		    (regexp
+		     ,(regexp-opt (directory-files tramp-archive-test-archive)))
+		    (? " ->" (+ nonl)) "\n"))))))
 	  ;; Check error case.
 	  (with-temp-buffer
 	    (should-error
 	     (insert-directory
 	      (expand-file-name "baz" tramp-archive-test-archive) nil)
-	     :type tramp-file-missing)))
+	     :type 'file-missing)))
 
       ;; Cleanup.
       (tramp-archive-cleanup-hash))))
@@ -715,7 +708,7 @@ This tests also `access-file', `file-readable-p' and `file-regular-p'."
 	  ;; Check error case.
 	  (should-error
 	   (access-file tmp-name4  "error")
-	   :type tramp-file-missing))
+	   :type 'file-missing))
 
       ;; Cleanup.
       (tramp-archive-cleanup-hash))))
@@ -739,7 +732,7 @@ This tests also `access-file', `file-readable-p' and `file-regular-p'."
 	  (setq attr (directory-files-and-attributes tmp-name 'full))
 	  (dolist (elt attr)
 	    (should (equal (file-attributes (car elt)) (cdr elt))))
-	  (setq attr (directory-files-and-attributes tmp-name nil "\\`b"))
+	  (setq attr (directory-files-and-attributes tmp-name nil (rx bos "b")))
 	  (should (equal (mapcar #'car attr) '("bar"))))
 
       ;; Cleanup.
@@ -854,38 +847,27 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
       ;; Cleanup.
       (tramp-archive-cleanup-hash))))
 
-;; The functions were introduced in Emacs 26.1.
 (ert-deftest tramp-archive-test40-make-nearby-temp-file ()
   "Check `make-nearby-temp-file' and `temporary-file-directory'."
   (skip-unless tramp-archive-enabled)
-  ;; Since Emacs 26.1.
-  (skip-unless
-   (and (fboundp 'make-nearby-temp-file) (fboundp 'temporary-file-directory)))
 
-  ;; `make-nearby-temp-file' and `temporary-file-directory' exists
-  ;; since Emacs 26.1.  We don't want to see compiler warnings for
-  ;; older Emacsen.
   (let ((default-directory tramp-archive-test-archive)
 	tmp-file)
     ;; The file archive shall know a temporary file directory.  It is
     ;; not in the archive itself.
-    (should
-     (stringp (with-no-warnings (with-no-warnings (temporary-file-directory)))))
-    (should-not
-     (tramp-archive-file-name-p (with-no-warnings (temporary-file-directory))))
+    (should (stringp (temporary-file-directory)))
+    (should-not (tramp-archive-file-name-p (temporary-file-directory)))
 
     ;; A temporary file or directory shall not be located in the
     ;; archive itself.
-    (setq tmp-file
-	  (with-no-warnings (make-nearby-temp-file "tramp-archive-test")))
+    (setq tmp-file (make-nearby-temp-file "tramp-archive-test"))
     (should (file-exists-p tmp-file))
     (should (file-regular-p tmp-file))
     (should-not (tramp-archive-file-name-p tmp-file))
     (delete-file tmp-file)
     (should-not (file-exists-p tmp-file))
 
-    (setq tmp-file
-	  (with-no-warnings (make-nearby-temp-file "tramp-archive-test" 'dir)))
+    (setq tmp-file (make-nearby-temp-file "tramp-archive-test" 'dir))
     (should (file-exists-p tmp-file))
     (should (file-directory-p tmp-file))
     (should-not (tramp-archive-file-name-p tmp-file))
@@ -909,7 +891,7 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
 		 (zerop (nth 1 fsi))
 		 (zerop (nth 2 fsi))))))
 
-(ert-deftest tramp-archive-test45-auto-load ()
+(ert-deftest tramp-archive-test47-auto-load ()
   "Check that `tramp-archive' autoloads properly."
   :tags '(:expensive-test)
   (skip-unless tramp-archive-enabled)
@@ -937,11 +919,15 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
 	(dolist (file `("/mock::foo" ,(concat tramp-archive-test-archive "foo")))
           (should
            (string-match
-	    (format
-	     "tramp-archive loaded: %s[[:ascii:]]+tramp-archive loaded: %s"
-	     (tramp-archive-file-name-p default-directory)
-	     (or (tramp-archive-file-name-p default-directory)
-		 (and enabled (tramp-archive-file-name-p file))))
+	    (tramp-compat-rx
+	     "tramp-archive loaded: "
+	     (literal (symbol-name
+		       (tramp-archive-file-name-p default-directory)))
+	     (+ ascii)
+	     "tramp-archive loaded: "
+	     (literal (symbol-name
+		       (or (tramp-archive-file-name-p default-directory)
+			   (and enabled (tramp-archive-file-name-p file))))))
 	    (shell-command-to-string
 	     (format
 	      "%s -batch -Q -L %s --eval %s --eval %s"
@@ -952,7 +938,7 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
 	       (format "(setq tramp-archive-enabled %s)" enabled))
 	      (shell-quote-argument (format code file)))))))))))
 
-(ert-deftest tramp-archive-test45-delay-load ()
+(ert-deftest tramp-archive-test47-delay-load ()
   "Check that `tramp-archive' is loaded lazily, only when needed."
   :tags '(:expensive-test)
   (skip-unless tramp-archive-enabled)
@@ -978,9 +964,10 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
     (dolist (tae '(t nil))
       (should
        (string-match
-	(format
-	 "tramp-archive loaded: nil[[:ascii:]]+tramp-archive loaded: nil[[:ascii:]]+tramp-archive loaded: %s"
-	 tae)
+	(tramp-compat-rx
+	 "tramp-archive loaded: nil" (+ ascii)
+	 "tramp-archive loaded: nil" (+ ascii)
+	 "tramp-archive loaded: " (literal (symbol-name tae)))
         (shell-command-to-string
          (format
 	  "%s -batch -Q -L %s --eval %s"
@@ -1028,7 +1015,8 @@ This tests also `file-executable-p', `file-writable-p' and `set-file-modes'."
 	     (apply
 	      'append
 	      (mapcar
-	       (lambda (x) (directory-files (concat dir x) 'full "uu\\'" 'sort))
+	       (lambda (x)
+		 (directory-files (concat dir x) 'full (rx "uu" eos) 'sort))
 	       '("~/src/libarchive-3.2.2/libarchive/test"
 		 "~/src/libarchive-3.2.2/cpio/test"
 		 "~/src/libarchive-3.2.2/tar/test"))))
